import pickle

import abjad

import nauert


def test_QuantizationJob_pickle_01():
    job_id = 1
    definition = {2: {2: {2: None}, 3: None}, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    q_event_proxies = [
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(0), ["A"], index=1),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(1, 5), ["B"], index=2),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(1, 4), ["C"], index=3),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(1, 3), ["D"], index=4),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(2, 5), ["E"], index=5),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(1, 2), ["F"], index=6),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(3, 5), ["G"], index=7),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(2, 3), ["H"], index=8),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(3, 4), ["I"], index=9),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(4, 5), ["J"], index=10),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
        nauert.QEventProxy(
            nauert.SilentQEvent(abjad.Offset(1), ["K"], index=11),
            abjad.Offset(0),
            abjad.Offset(1),
        ),
    ]
    job = nauert.QuantizationJob(job_id, search_tree, q_event_proxies)
    pickled = pickle.loads(pickle.dumps(job))
    assert pickled is not job
    assert repr(pickled) == repr(job)

    job()
    pickled = pickle.loads(pickle.dumps(job))
    assert pickled is not job
    assert repr(pickled) == repr(job)



import abjad

from . import attackpointoptimizers as _attackpointoptimizers
from . import gracehandlers as _gracehandlers
from . import heuristics as _heuristics
from . import jobhandlers as _jobhandlers
from . import qeventsequence as _qeventsequence
from . import qschemas as _qschemas


def quantize(
    q_event_sequence: _qeventsequence.QEventSequence,
    q_schema: _qschemas.QSchema | None = None,
    grace_handler: _gracehandlers.GraceHandler | None = None,
    heuristic: _heuristics.Heuristic | None = None,
    job_handler: _jobhandlers.JobHandler | None = None,
    attack_point_optimizer: _attackpointoptimizers.AttackPointOptimizer | None = None,
    attach_tempos: bool = True,
) -> abjad.Voice:
    r"""
    Quantizer function.

    Quantizes sequences of attack-points, encapsulated by ``QEventSequences``,
    into score trees.

    ..  container:: example

        >>> durations = [1000] * 8
        >>> pitches = range(8)
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> method = nauert.QEventSequence.from_millisecond_pitch_pairs
        >>> q_event_sequence = method(pairs)

    ..  container:: example

        Quantization defaults to outputting into a 4/4, quarter=60 musical structure:

        >>> result = nauert.quantize(q_event_sequence)
        >>> staff = abjad.Staff([result])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \new Voice
                    {
                        {
                            \tempo 4=60
                            \time 4/4
                            c'4
                            cs'4
                            d'4
                            ef'4
                        }
                        {
                            e'4
                            f'4
                            fs'4
                            g'4
                        }
                    }
                }
            >>

    ..  container:: example

        However, the behavior of the ``quantize`` function can be modified at
        call-time. Passing a ``QSchema`` instance will alter the
        macro-structure of the output.

        Here, we quantize using settings specified by a ``MeasurewiseQSchema``,
        which will cause the ``quantize`` function to group the output into
        measures with different tempi and time signatures:

        >>> measurewise_q_schema = nauert.MeasurewiseQSchema(
        ...     {
        ...         "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 78),
        ...         "time_signature": abjad.TimeSignature((2, 4)),
        ...     },
        ...     {
        ...         "tempo": abjad.MetronomeMark(abjad.Duration(1, 8), 57),
        ...         "time_signature": abjad.TimeSignature((5, 4)),
        ...     },
        ... )
        >>> result = nauert.quantize(
        ...     q_event_sequence,
        ...     q_schema=measurewise_q_schema,
        ... )
        >>> staff = abjad.Staff([result])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \new Voice
                    {
                        {
                            \tempo 4=78
                            \time 2/4
                            c'4
                            ~
                            \tuplet 5/4 {
                                c'16.
                                cs'8..
                                ~
                            }
                        }
                        {
                            \tuplet 7/4 {
                                \tempo 8=57
                                \time 5/4
                                cs'16.
                                d'8
                                ~
                            }
                            \tuplet 5/4 {
                                d'16
                                ef'16.
                                ~
                            }
                            \tuplet 3/2 {
                                ef'16
                                e'8
                                ~
                            }
                            \tuplet 7/4 {
                                e'16
                                f'8
                                ~
                                f'32
                                ~
                            }
                            f'32
                            fs'16.
                            ~
                            \tuplet 5/4 {
                                fs'32
                                g'8
                                ~
                            }
                            \tuplet 7/4 {
                                g'32
                                r32
                                r16
                                r16
                                r16
                                r16
                                r16
                                r16
                            }
                            r4
                        }
                    }
                }
            >>

    ..  container:: example

        Here we quantize using settings specified by a ``BeatwiseQSchema``,
        which keeps the output of the ``quantize`` function "flattened",
        without measures or explicit time signatures.  The default beat-wise
        settings of quarter=60 persists until the third "beatspan":

        >>> beatwise_q_schema = nauert.BeatwiseQSchema(
        ... {
        ...     2: {"tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 120)},
        ...     5: {"tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 90)},
        ...     7: {"tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 30)},
        ... })

        >>> result = nauert.quantize(
        ...     q_event_sequence,
        ...     q_schema=beatwise_q_schema,
        ... )
        >>> staff = abjad.Staff([result])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    \new Voice
                    {
                        \tempo 4=60
                        c'4
                        cs'4
                        \tempo 4=120
                        d'2
                        ef'4
                        ~
                        \tempo 4=90
                        ef'8.
                        e'4
                        ~
                        e'16
                        ~
                        \tuplet 3/2 {
                            \tempo 4=30
                            e'32
                            f'8.
                            fs'8
                            ~
                            fs'32
                            ~
                        }
                        \tuplet 3/2 {
                            fs'32
                            g'8.
                            r32
                            r8
                        }
                    }
                }
            >>

        Note that ``TieChains`` are generally fused together in the above
        example, but break at tempo changes.

    ..  container::

        The use of `BeatwiseQSchema` and `MeasurewiseAttackPointOptimizer` is
        not supported. Please raise an issue if you would like this to be
        supported in the future.

        >>> q_schema = nauert.BeatwiseQSchema()
        >>> attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
        >>> result = nauert.quantize(
        ...     q_event_sequence,
        ...     attack_point_optimizer=attack_point_optimizer,
        ...     q_schema=q_schema,
        ... )
        Traceback (most recent call last):
            ...
        TypeError: BeatwiseQTarget is not ... with MeasurewiseAttackPointOptimizer.

    Other keyword arguments are:

        * ``grace_handler``: a ``GraceHandler`` instance controls whether and
          how grace notes are used in the output.  Options currently include
          ``CollapsingGraceHandler``, ``ConcatenatingGraceHandler`` and
          ``DiscardingGraceHandler``.

        * ``heuristic``: a ``Heuristic`` instance controls how output rhythms
          are selected from a pool of candidates.  Options currently include
          the ``DistanceHeuristic`` class.

        * ``job_handler``: a ``JobHandler`` instance controls whether or not
          parallel processing is used during the quantization process.
          Options include the ``SerialJobHandler`` and ``ParallelJobHandler``
          classes.

        * ``attack_point_optimizer``: an ``AttackPointOptimizer`` instance
          controls whether and how logical ties are re-notated.
          Options currently include ``MeasurewiseAttackPointOptimizer``,
          ``NaiveAttackPointOptimizer`` and ``NullAttackPointOptimizer``.

    Refer to the reference pages for ``BeatwiseQSchema`` and
    ``MeasurewiseQSchema`` for more information on controlling the ``quantize``
    function's output, and to the reference on ``SearchTree`` for information
    on controlling the rhythmic complexity of that same output.
    """
    # TODO: assert isinstance(q_event_sequence, QEventSequence)
    q_event_sequence = _qeventsequence.QEventSequence(q_event_sequence)
    if q_schema is None:
        q_schema = _qschemas.MeasurewiseQSchema()
    assert isinstance(q_schema, _qschemas.QSchema)
    q_target = q_schema(q_event_sequence.duration_in_ms)
    notation = q_target(
        q_event_sequence,
        grace_handler=grace_handler,
        heuristic=heuristic,
        job_handler=job_handler,
        attack_point_optimizer=attack_point_optimizer,
        attach_tempos=attach_tempos,
    )
    return notation



import abjad

import nauert


def test_MeasurewiseAttackPointOptimizer___call___01():
    string = r"""
        \tuplet 11/8
        {
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            \grace {
                e''16
                e''16
            }
            d''8
            ~
            d''8
            r8
            r8
        }
        """
    string = abjad.string.normalize(string)
    container = abjad.Container(string)
    time_signature = abjad.TimeSignature((4, 4))
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    attack_point_optimizer(container, time_signature)
    string = abjad.lilypond(container)
    assert string == abjad.string.normalize(
        r"""
        {
            \tuplet 11/8
            {
                a''2..
                \grace {
                    e''16
                    e''16
                }
                d''4
                r4
            }
        }
        """
    ), print(string)


def test_annotations_survive_measurewise_attack_point_optimimzer():
    container = abjad.Container("c'8~ c'8 c'8 c'4. c'4")
    abjad.annotate(container[0], "q_event_attachments", "test_indicator_0")
    assert (
        abjad.get.annotation(container[0], "q_event_attachments") == "test_indicator_0"
    )
    abjad.annotate(container[3], "q_event_attachments", "test_indicator_1")
    assert (
        abjad.get.annotation(container[3], "q_event_attachments") == "test_indicator_1"
    )
    time_signature = abjad.TimeSignature((4, 4))
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    attack_point_optimizer(container, time_signature)
    string = abjad.lilypond(container)
    assert string == abjad.string.normalize(
        r"""
        {
            c'4
            c'8
            c'8
            ~
            c'4
            c'4
        }
        """
    )
    assert (
        abjad.get.annotation(container[0], "q_event_attachments") == "test_indicator_0"
    )
    assert (
        abjad.get.annotation(container[2], "q_event_attachments") == "test_indicator_1"
    )


def test_annotations_survive_naive_attack_point_optimimzer():
    container = abjad.Container("c'8~ c'8 c'8 c'4. c'4")
    abjad.annotate(container[0], "test_annotation", "test_indicator")
    assert abjad.get.annotation(container[0], "test_annotation") == "test_indicator"
    attack_point_optimizer = nauert.NaiveAttackPointOptimizer()
    attack_point_optimizer(container)
    string = abjad.lilypond(container)
    assert string == abjad.string.normalize(
        r"""
        {
            c'4
            c'8
            c'4.
            c'4
        }
        """
    )
    assert abjad.get.annotation(container[0], "test_annotation") == "test_indicator"



import abjad

import nauert


def test_DistanceHeuristic___call___01():
    heuristic = nauert.DistanceHeuristic()
    q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0, 1])
    q_event_b = nauert.SilentQEvent(abjad.Offset(500))
    q_event_c = nauert.PitchedQEvent(abjad.Offset(750), [3, 7])
    proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
    proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.5))
    proxy_c = nauert.QEventProxy(q_event_c, abjad.Offset(0.75))
    definition = {2: {2: None}, 3: None, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    job = nauert.QuantizationJob(1, search_tree, [proxy_a, proxy_b, proxy_c])
    job()
    q_target_beat = nauert.QTargetBeat()
    q_target_beat._q_grids = job.q_grids
    q_target_beats = heuristic((q_target_beat,))
    q_target_beat = q_target_beats[0]
    q_grid = q_target_beat.q_grid
    assert q_grid.distance == 0
    rtm = q_grid.rtm_format
    assert rtm == "(1 ((1 (1 1)) (1 (1 1))))"


def test_DistanceHeuristic___call___02():
    heuristic = nauert.DistanceHeuristic()
    q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0, 1])
    q_event_b = nauert.SilentQEvent(abjad.Offset(500))
    q_event_c = nauert.PitchedQEvent(abjad.Offset(750), [3, 7])
    proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
    proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.5))
    proxy_c = nauert.QEventProxy(q_event_c, abjad.Offset(0.75))
    definition = {2: None, 3: None, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    job = nauert.QuantizationJob(1, search_tree, [proxy_a, proxy_b, proxy_c])
    job()
    q_target_beat = nauert.QTargetBeat()
    q_target_beat._q_grids = job.q_grids
    q_target_beats = heuristic((q_target_beat,))
    q_target_beat = q_target_beats[0]
    q_grid = q_target_beat.q_grid
    assert q_grid.distance == abjad.Duration(1, 15)
    rtm = q_grid.rtm_format
    assert rtm == "(1 (1 1 1 1 1))"



import abc
import bisect
import copy
import typing

import abjad

from . import attackpointoptimizers as _attackpointoptimizers
from . import gracehandlers as _gracehandlers
from . import heuristics as _heuristics
from . import jobhandlers as _jobhandlers
from . import qeventproxy as _qeventproxy
from . import qevents as _qevents
from . import qeventsequence as _qeventsequence
from . import qtargetitems as _qtargetitems


class QTarget(abc.ABC):
    """
    Q-target.

    ``QTarget`` is created by a concrete ``QSchema`` instance.

    ``QTarget`` represents the mold into which the timepoints contained by a
    ``QSequence`` instance will be poured, as structured by that ``QSchema``
    instance.

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_items",)

    ### INITIALIZATION ###

    def __init__(self, items: typing.Sequence[_qtargetitems.QTargetItem] | None = None):
        items = [] if items is None else items
        assert all(isinstance(x, self.item_class) for x in items)
        self._items: typing.Sequence[_qtargetitems.QTargetItem] = ()
        if len(items) > 0:
            self._items = tuple(sorted(items, key=lambda x: x.offset_in_ms))

    ### SPECIAL METHODS ###

    # TODO: typehint
    def __call__(
        self,
        q_event_sequence: _qeventsequence.QEventSequence,
        grace_handler: _gracehandlers.GraceHandler | None = None,
        heuristic: _heuristics.Heuristic | None = None,
        job_handler: _jobhandlers.JobHandler | None = None,
        attack_point_optimizer: (
            _attackpointoptimizers.AttackPointOptimizer | None
        ) = None,
        attach_tempos: bool = True,
    ):
        """
        Calls q-target.
        """
        assert isinstance(q_event_sequence, _qeventsequence.QEventSequence)
        if grace_handler is None:
            grace_handler = _gracehandlers.ConcatenatingGraceHandler()
        assert isinstance(grace_handler, _gracehandlers.GraceHandler)
        if heuristic is None:
            heuristic = _heuristics.DistanceHeuristic()
        assert isinstance(heuristic, _heuristics.Heuristic)
        if job_handler is None:
            job_handler = _jobhandlers.SerialJobHandler()
        assert isinstance(job_handler, _jobhandlers.JobHandler)
        if attack_point_optimizer is None:
            attack_point_optimizer = _attackpointoptimizers.NaiveAttackPointOptimizer()
        assert isinstance(
            attack_point_optimizer, _attackpointoptimizers.AttackPointOptimizer
        )
        if isinstance(self, BeatwiseQTarget) and isinstance(
            attack_point_optimizer,
            _attackpointoptimizers.MeasurewiseAttackPointOptimizer,
        ):
            message = "{} is not supposed to be used together with {}.".format(
                self.__class__.__name__, attack_point_optimizer.__class__.__name__
            )
            raise TypeError(message)
        # parcel QEvents out to each beat
        beats = self.beats
        offsets = sorted([beat.offset_in_ms for beat in beats])
        for q_event in q_event_sequence:
            index = bisect.bisect(offsets, q_event.offset) - 1
            beat = beats[index]
            beat.q_events.append(q_event)
        # generate QuantizationJobs and process with the JobHandler
        jobs = [beat(i) for i, beat in enumerate(beats)]
        jobs = [job for job in jobs if job]
        jobs = job_handler(jobs)
        for job in jobs:
            assert job is not None
            beats[job.job_id]._q_grids = job.q_grids
        # select the best QGrid for each beat, according to the Heuristic
        beats = heuristic(beats)
        # shift QEvents attached to each QGrid's "next downbeat"
        # over to the next QGrid's first leaf - the real downbeat
        orphaned_q_events_proxies = self._shift_downbeat_q_events_to_next_q_grid()
        # TODO: handle a final QGrid with QEvents attached to its next_downbeat
        # TODO: remove a final QGrid with no QEvents
        self._regroup_q_grid_with_unnecessary_divisions()
        # convert the QGrid representation into notation,
        # handling grace-note behavior with the GraceHandler
        notation = self._notate(
            attach_tempos=attach_tempos,
            attack_point_optimizer=attack_point_optimizer,
            grace_handler=grace_handler,
        )
        handle_orphaned_q_events = getattr(
            grace_handler, "handle_orphaned_q_event_proxies", None
        )
        if callable(handle_orphaned_q_events) and orphaned_q_events_proxies:
            last_leaf = abjad.get.leaf(notation, -1)
            handle_orphaned_q_events(last_leaf, orphaned_q_events_proxies)
        return notation

    ### PRIVATE METHODS ###

    @abc.abstractmethod
    def _notate(
        self,
        grace_handler: _gracehandlers.GraceHandler,
        attack_point_optimizer: _attackpointoptimizers.AttackPointOptimizer,
        attach_tempos: bool = True,
    ) -> abjad.Voice:
        raise NotImplementedError

    def _notate_leaves(
        self,
        grace_handler: _gracehandlers.GraceHandler,
        voice: abjad.Voice | None = None,
    ):
        for leaf in abjad.iterate.leaves(voice):
            if leaf._has_indicator(dict):
                annotation = leaf._get_indicator(dict)
                q_events = annotation["q_events"]
                pitches, attachments, grace_container = grace_handler(q_events)
                new_leaf: abjad.Leaf
                if not pitches:
                    new_leaf = abjad.Rest(leaf)
                elif 1 < len(pitches):
                    new_leaf = abjad.Chord(leaf)
                    new_leaf.written_pitches = pitches
                else:
                    new_leaf = abjad.Note(leaf)
                    new_leaf.written_pitch = pitches[0]
                if attachments is not None:
                    abjad.annotate(new_leaf, "q_event_attachments", attachments)
                if grace_container:
                    abjad.attach(grace_container, new_leaf)
                abjad.mutate.replace(leaf, new_leaf)
                if not isinstance(new_leaf, abjad.Rest):
                    abjad.annotate(new_leaf, "tie_to_next", True)
                elif abjad.get.indicator(new_leaf, abjad.Tie):
                    abjad.detach(abjad.Tie, new_leaf)
            else:
                previous_leaf = abjad._iterlib._get_leaf(leaf, -1)
                if isinstance(previous_leaf, abjad.Rest):
                    new_leaf = type(previous_leaf)(leaf.written_duration)
                elif isinstance(previous_leaf, abjad.Note):
                    new_leaf = type(previous_leaf)(
                        previous_leaf.written_pitch, leaf.written_duration
                    )
                else:
                    new_leaf = type(previous_leaf)(
                        previous_leaf.written_pitches, leaf.written_duration
                    )
                abjad.mutate.replace(leaf, new_leaf)
                if abjad.get.annotation(previous_leaf, "tie_to_next") is True:
                    leaves = [previous_leaf, new_leaf]
                    abjad.tie(leaves)
                    abjad.annotate(new_leaf, "tie_to_next", True)
            if leaf._has_indicator(abjad.MetronomeMark):
                tempo = leaf._get_indicator(abjad.MetronomeMark)
                abjad.detach(abjad.MetronomeMark, leaf)
                abjad.detach(abjad.MetronomeMark, new_leaf)
                abjad.attach(tempo, new_leaf)
            if leaf._has_indicator(abjad.TimeSignature):
                time_signature = leaf._get_indicator(abjad.TimeSignature)
                abjad.detach(abjad.TimeSignature, leaf)
                abjad.detach(abjad.TimeSignature, new_leaf)
                abjad.attach(time_signature, new_leaf)

    def _regroup_q_grid_with_unnecessary_divisions(self):
        for beat in self.beats:
            beat.q_grid.regroup_leaves_with_unencessary_divisions()

    def _shift_downbeat_q_events_to_next_q_grid(self) -> list[_qeventproxy.QEventProxy]:
        beats = self.beats
        assert beats[-1].q_grid is not None
        for one, two in abjad.sequence.nwise(beats):
            one_q_events = one.q_grid.next_downbeat.q_event_proxies
            two_q_events = two.q_grid.leaves[0].q_event_proxies
            while one_q_events:
                two_q_events.insert(0, one_q_events.pop())
        return [
            proxy
            for proxy in beats[-1].q_grid.next_downbeat.q_event_proxies
            if not isinstance(proxy.q_event, _qevents.TerminalQEvent)
        ]

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def beats(self) -> tuple[_qtargetitems.QTargetBeat, ...]:
        """
        Gets beats of q-target.
        """
        raise NotImplementedError

    @property
    def duration_in_ms(self) -> abjad.Duration:
        """
        Gets duration of q-target in milliseconds.
        """
        last_item = self._items[-1]
        return last_item.offset_in_ms + last_item.duration_in_ms

    @abc.abstractproperty
    def item_class(self):
        """
        Gets item class of q-target.
        """
        raise NotImplementedError

    @property
    def items(self):
        """
        Gets items of q-target.
        """
        return self._items


class BeatwiseQTarget(QTarget):
    """
    Beatwise q-target.

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZATION ###

    def __init__(self, items: typing.Sequence[_qtargetitems.QTargetBeat] | None = None):
        items = [] if items is None else items
        self._items: typing.Sequence[_qtargetitems.QTargetBeat]
        super().__init__(items)

    ### PRIVATE METHODS ###

    def _notate(
        self,
        grace_handler: _gracehandlers.GraceHandler,
        attack_point_optimizer: _attackpointoptimizers.AttackPointOptimizer,
        attach_tempos: bool = True,
    ) -> abjad.Voice:
        voice = abjad.Voice()
        # generate the first
        beat = self._items[0]
        assert isinstance(beat, _qtargetitems.QTargetBeat) and beat.q_grid is not None
        components = beat.q_grid(beat.beatspan)
        voice.extend(components)
        if attach_tempos:
            attachment_target: abjad.Component = components[0]
            leaves = abjad.select.leaves(attachment_target)
            if isinstance(attachment_target, abjad.Container):
                attachment_target = leaves[0]
            tempo = copy.deepcopy(beat.tempo)
            abjad.attach(tempo, attachment_target)
        # generate the rest pairwise, comparing tempi
        for beat_one, beat_two in abjad.sequence.nwise(self.items):
            components = beat_two.q_grid(beat_two.beatspan)
            voice.extend(components)
            if (beat_two.tempo != beat_one.tempo) and attach_tempos:
                attachment_target = components[0]
                leaves = abjad.select.leaves(attachment_target)
                if isinstance(attachment_target, abjad.Container):
                    attachment_target = leaves[0]
                tempo = copy.deepcopy(beat_two.tempo)
                abjad.attach(tempo, attachment_target)
        # apply logical ties, pitches, grace containers
        self._notate_leaves(grace_handler=grace_handler, voice=voice)
        # partition logical ties in voice
        attack_point_optimizer(voice)
        return voice

    ### PUBLIC PROPERTIES ###

    @property
    def beats(self) -> tuple[_qtargetitems.QTargetBeat, ...]:
        """
        Gets beats of beatwise q-target.
        """
        return tuple(self._items)

    @property
    def item_class(self) -> type[_qtargetitems.QTargetBeat]:
        """
        Gets item class of beatwise q-target.
        """
        return _qtargetitems.QTargetBeat


class MeasurewiseQTarget(QTarget):
    """
    Measurewise quantization target.

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZATION ###

    def __init__(
        self, items: typing.Sequence[_qtargetitems.QTargetMeasure] | None = None
    ):
        super().__init__(items)

    ### PRIVATE METHODS ###

    def _notate(
        self,
        grace_handler: _gracehandlers.GraceHandler,
        attack_point_optimizer: _attackpointoptimizers.AttackPointOptimizer,
        attach_tempos: bool = True,
    ) -> abjad.Voice:
        voice = abjad.Voice()
        # generate the first
        q_target_measure = self._items[0]
        assert isinstance(q_target_measure, _qtargetitems.QTargetMeasure)
        time_signature = q_target_measure.time_signature
        measure = abjad.Container()
        for beat in q_target_measure.beats:
            components = beat.q_grid(beat.beatspan)
            measure.extend(components)
        voice.append(measure)
        leaf = abjad.get.leaf(measure, 0)
        abjad.attach(time_signature, leaf)
        if attach_tempos:
            tempo = copy.deepcopy(q_target_measure.tempo)
            leaf = abjad.get.leaf(measure, 0)
            abjad.attach(tempo, leaf)
        # generate the rest pairwise, comparing tempi
        pairs = abjad.sequence.nwise(self.items)
        for q_target_measure_one, q_target_measure_two in pairs:
            measure = abjad.Container()
            voice.append(measure)
            for beat in q_target_measure_two.beats:
                measure.extend(beat.q_grid(beat.beatspan))
            if (
                q_target_measure_two.time_signature
                != q_target_measure_one.time_signature
            ):
                time_signature = q_target_measure_two.time_signature
                leaf = abjad.get.leaf(measure, 0)
                abjad.attach(time_signature, leaf)
            if (
                q_target_measure_two.tempo != q_target_measure_one.tempo
            ) and attach_tempos:
                tempo = copy.deepcopy(q_target_measure_two.tempo)
                leaf = abjad.get.leaf(measure, 0)
                abjad.attach(tempo, leaf)
        # apply logical ties, pitches, grace containers
        self._notate_leaves(grace_handler=grace_handler, voice=voice)
        # partition logical ties in each measure
        for index, measure in enumerate(voice):
            if isinstance(
                attack_point_optimizer,
                _attackpointoptimizers.MeasurewiseAttackPointOptimizer,
            ):
                # then we need to pass the time signature of each measure
                attack_point_optimizer(measure, self.items[index].time_signature)
            else:
                attack_point_optimizer(measure)
        return voice

    ### PUBLIC PROPERTIES ###

    @property
    def beats(self) -> tuple[_qtargetitems.QTargetBeat, ...]:
        """
        Gets beats of measurewise q-target.
        """
        return tuple([beat for item in self.items for beat in item.beats])

    @property
    def item_class(self) -> type[_qtargetitems.QTargetMeasure]:
        """
        Gets item class of measurewise q-target.
        """
        return _qtargetitems.QTargetMeasure



import abjad
import pytest

import nauert


def test_BeatwiseQSchemaItem___new___01():
    item = nauert.BeatwiseQSchemaItem()
    assert item.beatspan is None
    assert item.search_tree is None
    assert item.tempo is None


def test_BeatwiseQSchemaItem___new___02():
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
    item = nauert.BeatwiseQSchemaItem(tempo=metronome_mark)
    assert item.beatspan is None
    assert item.search_tree is None
    assert item.tempo == metronome_mark


def test_BeatwiseQSchemaItem___new___03():
    item = nauert.BeatwiseQSchemaItem(beatspan=abjad.Duration(1, 8))
    assert item.beatspan == abjad.Duration(1, 8)
    assert item.search_tree is None
    assert item.tempo is None


def test_BeatwiseQSchemaItem___new___04():
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 57)
    item = nauert.BeatwiseQSchemaItem(
        beatspan=abjad.Duration(1, 8), tempo=metronome_mark
    )
    assert item.beatspan == abjad.Duration(1, 8)
    assert item.search_tree is None
    assert item.tempo == metronome_mark


def test_BeatwiseQSchemaItem___new___05():
    tempo = abjad.MetronomeMark(textual_indication="lento")
    with pytest.raises(AssertionError):
        nauert.BeatwiseQSchemaItem(tempo=tempo)



import abjad

import nauert


def test_QEventSequence_from_millisecond_pitch_pairs_01():
    durations = [100, 200, 100, 300, 350, 400, 600]
    pitches = [0, None, None, [1, 4], None, 5, 7]
    pairs = tuple(zip(durations, pitches))
    q_events = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(100, 1)),
            nauert.PitchedQEvent(
                abjad.Offset(400, 1),
                (abjad.NamedPitch("cs'"), abjad.NamedPitch("e'")),
            ),
            nauert.SilentQEvent(abjad.Offset(700, 1)),
            nauert.PitchedQEvent(abjad.Offset(1050, 1), (abjad.NamedPitch("f'"),)),
            nauert.PitchedQEvent(abjad.Offset(1450, 1), (abjad.NamedPitch("g'"),)),
            nauert.TerminalQEvent(abjad.Offset(2050, 1)),
        )
    )



import abjad

import nauert


def test_ConcatenatingGraceHandler___init___01():
    grace_handler = nauert.ConcatenatingGraceHandler()
    assert grace_handler.grace_duration == abjad.Duration(1, 16)
    assert grace_handler.discard_grace_rest is True


def test_ConcatenatingGraceHandler___init___02():
    grace_handler = nauert.ConcatenatingGraceHandler(discard_grace_rest=False)
    assert grace_handler.grace_duration == abjad.Duration(1, 16)
    assert grace_handler.discard_grace_rest is False
    assert grace_handler.replace_rest_with_final_grace_note is True


def test_ConcatenatingGraceHandler___init___03():
    duration = abjad.Duration(1, 32)
    grace_handler = nauert.ConcatenatingGraceHandler(grace_duration=duration)
    assert grace_handler.grace_duration == abjad.Duration(1, 32)
    assert grace_handler.discard_grace_rest is True
    assert grace_handler.replace_rest_with_final_grace_note is True


def test_ConcatenatingGraceHandler___init___04():
    grace_handler = nauert.ConcatenatingGraceHandler(
        replace_rest_with_final_grace_note=True
    )
    assert grace_handler.grace_duration == abjad.Duration(1, 16)
    assert grace_handler.discard_grace_rest is True
    assert grace_handler.replace_rest_with_final_grace_note is True



import abc
import typing

import abjad

from . import searchtrees as _searchtrees


class QSchemaItem(abc.ABC):
    """
    Abstract q-schema item.

    Represents a change of state in the timeline of a quantization process.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_search_tree", "_tempo")

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(
        self,
        search_tree: _searchtrees.SearchTree | None = None,
        tempo: abjad.MetronomeMark | None = None,
    ) -> None:
        if search_tree is not None:
            assert isinstance(search_tree, _searchtrees.SearchTree)
        self._search_tree = search_tree
        if tempo is not None:
            assert isinstance(tempo, abjad.MetronomeMark), repr(tempo)
            assert not tempo.is_imprecise
        self._tempo = tempo

    ### PUBLIC PROPERTIES ###

    @property
    def search_tree(self) -> typing.Optional[_searchtrees.SearchTree]:
        """
        The optionally defined search tree.
        """
        return self._search_tree

    @property
    def tempo(self) -> typing.Optional[abjad.MetronomeMark]:
        """
        The optionally defined tempo.
        """
        return self._tempo


class BeatwiseQSchemaItem(QSchemaItem):
    """
    Beatwise q-schema item.

    ..  container:: example

        Represents a change of state in the timeline of an unmetered
        quantization process:

        >>> nauert.BeatwiseQSchemaItem()
        BeatwiseQSchemaItem(beatspan=None, search_tree=None, tempo=None)

    ..  container:: example

        Defines a change in tempo:

        >>> metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
        >>> nauert.BeatwiseQSchemaItem(tempo=metronome_mark)
        BeatwiseQSchemaItem(...)

    ..  container:: example

        Defines a change in beatspan:

        >>> nauert.BeatwiseQSchemaItem(beatspan=abjad.Duration(1, 8))
        BeatwiseQSchemaItem(beatspan=Duration(1, 8), search_tree=None, tempo=None)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_beatspan",)

    ### INITIALIZER ###

    def __init__(
        self,
        beatspan: abjad.Duration | None = None,
        search_tree: _searchtrees.SearchTree | None = None,
        tempo: abjad.MetronomeMark | None = None,
    ) -> None:
        if beatspan is not None:
            assert isinstance(beatspan, abjad.Duration), repr(beatspan)
        QSchemaItem.__init__(self, search_tree=search_tree, tempo=tempo)
        if beatspan is not None:
            beatspan = abjad.Duration(beatspan)
            assert 0 < beatspan
        self._beatspan = beatspan

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(beatspan={self.beatspan!r},"
        string += f" search_tree={self.search_tree!r}, tempo={self.tempo!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def beatspan(self) -> typing.Optional[abjad.Duration]:
        """
        The optionally defined beatspan duration.
        """
        return self._beatspan


class MeasurewiseQSchemaItem(QSchemaItem):
    r"""
    Measurewise q-schema item.

    Represents a change of state in the timeline of a metered quantization process.

    >>> q_schema_item = nauert.MeasurewiseQSchemaItem()

    ..  container:: example

        Defines a change in tempo:

        >>> metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
        >>> mark = nauert.MeasurewiseQSchemaItem(tempo=metronome_mark).tempo
        >>> abjad.lilypond(mark)
        '\\tempo 4=60'

    ..  container:: example

        Defines a change in time signature:

        >>> time_signature = abjad.TimeSignature((6, 8))
        >>> nauert.MeasurewiseQSchemaItem(time_signature=time_signature).time_signature
        TimeSignature(pair=(6, 8), hide=False, partial=None)

    ..  container:: example

        Tests for beatspan given a defined time signature:

        >>> time_signature = abjad.TimeSignature((6, 8))
        >>> nauert.MeasurewiseQSchemaItem(time_signature=time_signature).beatspan
        Duration(1, 8)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_time_signature", "_use_full_measure")

    ### INITIALIZER ###

    def __init__(
        self,
        search_tree: _searchtrees.SearchTree | None = None,
        tempo: abjad.MetronomeMark | None = None,
        time_signature: abjad.TimeSignature | None = None,
        use_full_measure: bool | None = None,
    ):
        if time_signature is not None:
            assert isinstance(time_signature, abjad.TimeSignature), repr(time_signature)
        QSchemaItem.__init__(self, search_tree=search_tree, tempo=tempo)
        self._time_signature: abjad.TimeSignature | None
        if isinstance(time_signature, abjad.TimeSignature):
            self._time_signature = abjad.TimeSignature(time_signature.pair)
        elif isinstance(time_signature, tuple):
            self._time_signature = abjad.TimeSignature(time_signature)
        else:
            assert time_signature is None
            self._time_signature = None
        if use_full_measure is not None:
            use_full_measure = bool(use_full_measure)
        self._use_full_measure = use_full_measure

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(search_tree={self.search_tree!r},"
        string += f" tempo={self.tempo!r}, time_signature={self.time_signature!r},"
        string += f" use_full_measure={self.use_full_measure!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def beatspan(self) -> typing.Optional[abjad.Duration]:
        """
        The beatspan duration, if a time signature was defined.
        """
        if self.time_signature is not None:
            if self.use_full_measure:
                return self.time_signature.duration
            else:
                return abjad.Duration(1, self.time_signature.denominator)
        return None

    @property
    def time_signature(self) -> typing.Optional[abjad.TimeSignature]:
        """
        The optionally defined time signature.
        """
        return self._time_signature

    @property
    def use_full_measure(self) -> typing.Optional[bool]:
        """
        If true, use the full measure as the beatspan.
        """
        return self._use_full_measure



import abc
import typing

import abjad

from . import qeventproxy as _qeventproxy
from . import qevents as _qevents
from . import qgrid as _qgrid
from . import quantizationjob as _quantizationjob
from . import searchtrees as _searchtrees


class QTargetItem(abc.ABC):
    """
    Abstract class for QTargetBeat and QTargetMeasure.
    """

    @abc.abstractproperty
    def offset_in_ms(self) -> abjad.Offset:
        raise NotImplementedError

    @abc.abstractproperty
    def duration_in_ms(self) -> abjad.Duration:
        raise NotImplementedError


class QTargetBeat(QTargetItem):
    """
    Q-target beat.

    Represents a single beat in a quantization target.

    ..  container:: example

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)
        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_beatspan",
        "_grouping",
        "_offset_in_ms",
        "_q_events",
        "_q_grid",
        "_q_grids",
        "_search_tree",
        "_tempo",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        beatspan: abjad.Duration = abjad.Duration(0),
        offset_in_ms: abjad.Offset = abjad.Offset(0),
        search_tree: _searchtrees.SearchTree | None = None,
        tempo: abjad.MetronomeMark = abjad.MetronomeMark(abjad.Duration(1, 4), 60),
    ):
        assert isinstance(beatspan, abjad.Duration), repr(beatspan)
        assert isinstance(offset_in_ms, abjad.Offset), repr(offset_in_ms)
        if search_tree is None:
            search_tree = _searchtrees.UnweightedSearchTree()
        assert isinstance(search_tree, _searchtrees.SearchTree)
        assert isinstance(tempo, abjad.MetronomeMark), repr(tempo)
        assert not tempo.is_imprecise
        q_events: list[_qevents.QEvent] = []
        q_grids: tuple[_qgrid.QGrid, ...] = ()
        self._beatspan = beatspan
        self._offset_in_ms = offset_in_ms
        self._q_events = q_events
        self._q_grid: _qgrid.QGrid | None = None
        self._q_grids = q_grids
        self._search_tree = search_tree
        self._tempo = tempo

    ### SPECIAL METHODS ###

    def __call__(
        self, job_id: int
    ) -> typing.Optional[_quantizationjob.QuantizationJob]:
        """
        Calls q-target beat.
        """
        if not self.q_events:
            return None
        assert all(isinstance(x, _qevents.QEvent) for x in self.q_events)
        q_event_proxies = []
        for q_event in self.q_events:
            q_event_proxy = _qeventproxy.QEventProxy(
                q_event,
                self.offset_in_ms,
                self.offset_in_ms + self.duration_in_ms,
            )
            q_event_proxies.append(q_event_proxy)
        return _quantizationjob.QuantizationJob(
            job_id, self.search_tree, q_event_proxies
        )

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(beatspan={self.beatspan!r},"
        string += f" offset_in_ms={self.offset_in_ms!r},"
        string += f" search_tree={self.search_tree!r}, tempo={self.tempo!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def beatspan(self) -> abjad.Duration:
        """
        Beatspan of q-target beat.

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)

        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

        >>> q_target_beat.beatspan
        Duration(1, 8)

        """
        return self._beatspan

    @property
    def duration_in_ms(self) -> abjad.Duration:
        """
        Duration in milliseconds of the q-target beat.

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)

        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

        >>> q_target_beat.duration_in_ms
        Duration(3750, 7)

        """
        return self.tempo.duration_to_milliseconds(self.beatspan)

    @property
    def offset_in_ms(self) -> abjad.Offset:
        """
        Offset in milliseconds of q-target beat.

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)

        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

        >>> q_target_beat.offset_in_ms
        Offset((1500, 1))

        """
        return self._offset_in_ms

    @property
    def q_events(self) -> list[_qevents.QEvent]:
        """
        A list for storing ``QEventProxy`` instances.

        Used internally by the ``quantize`` function.
        """
        return self._q_events

    @property
    def q_grid(self) -> typing.Optional[_qgrid.QGrid]:
        """
        The ``QGrid`` instance selected by a ``Heuristic``.

        Used internally by the ``quantize`` function.
        """
        return self._q_grid

    @property
    def q_grids(self) -> tuple[_qgrid.QGrid, ...]:
        """
        A tuple of ``QGrids`` generated by a ``QuantizationJob``.

        Used internally by the ``quantize`` function.
        """
        return self._q_grids

    @property
    def search_tree(
        self,
    ) -> _searchtrees.SearchTree | _searchtrees.UnweightedSearchTree:
        """
        Search tree of q-target beat.

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)

        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

        >>> q_target_beat.search_tree
        UnweightedSearchTree(definition={3: None})

        """
        return self._search_tree

    @property
    def tempo(self) -> abjad.MetronomeMark:
        r"""
        Gets tempo of q-target beat.

        >>> beatspan = abjad.Duration(1, 8)
        >>> offset_in_ms = abjad.Offset(1500)
        >>> search_tree = nauert.UnweightedSearchTree({3: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 56)
        >>> q_target_beat = nauert.QTargetBeat(
        ...     beatspan=beatspan,
        ...     offset_in_ms=offset_in_ms,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )
        >>> abjad.lilypond(q_target_beat.tempo)
        '\\tempo 4=56'

        """
        return self._tempo


class QTargetMeasure(QTargetItem):
    """
    Q-target measure.

    Represents a single measure in a measurewise quantization target.

    ..  container:: example

        >>> search_tree = nauert.UnweightedSearchTree({2: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
        >>> time_signature = abjad.TimeSignature((4, 4))
        >>> q_target_measure = nauert.QTargetMeasure(
        ...     offset_in_ms=abjad.Offset(1000),
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ...     time_signature=time_signature,
        ... )

    ..  container:: example

        ``QTargetMeasures`` group ``QTargetBeats``:

        >>> for q_target_beat in q_target_measure.beats:
        ...     print(q_target_beat.offset_in_ms, q_target_beat.duration_in_ms)
        1000 1000
        2000 1000
        3000 1000
        4000 1000

    ..  container:: example

        If ``use_full_measure`` is set, the ``QTargetMeasure`` will only ever
        contain a single ``QTargetBeat`` instance:

        >>> another_q_target_measure = nauert.QTargetMeasure(
        ...     offset_in_ms=abjad.Offset(1000),
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ...     time_signature=time_signature,
        ...     use_full_measure=True,
        ... )

        >>> for q_target_beat in another_q_target_measure.beats:
        ...     print(q_target_beat.offset_in_ms, q_target_beat.duration_in_ms)
        1000 4000

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_beats",
        "_offset_in_ms",
        "_search_tree",
        "_tempo",
        "_time_signature",
        "_use_full_measure",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        offset_in_ms: abjad.Offset = abjad.Offset(0),
        search_tree: _searchtrees.SearchTree | None = None,
        time_signature: abjad.TimeSignature = abjad.TimeSignature((4, 4)),
        tempo: abjad.MetronomeMark = abjad.MetronomeMark(abjad.Duration(4, 4)),
        use_full_measure: bool = False,
    ):
        assert isinstance(offset_in_ms, abjad.Offset), repr(offset_in_ms)
        if search_tree is None:
            search_tree = _searchtrees.UnweightedSearchTree()
        assert isinstance(search_tree, _searchtrees.SearchTree)
        assert isinstance(time_signature, abjad.TimeSignature), repr(time_signature)
        assert isinstance(tempo, abjad.MetronomeMark), repr(tempo)
        assert not tempo.is_imprecise
        _time_signature = abjad.TimeSignature(time_signature.pair)
        use_full_measure = bool(use_full_measure)
        beats = []
        if use_full_measure:
            beatspan = _time_signature.duration
            beat = QTargetBeat(
                beatspan=beatspan,
                offset_in_ms=offset_in_ms,
                search_tree=search_tree,
                tempo=tempo,
            )
            beats.append(beat)
        else:
            beatspan = abjad.Duration(1, _time_signature.denominator)
            current_offset_in_ms = offset_in_ms
            beatspan_duration_in_ms = tempo.duration_to_milliseconds(beatspan)
            for i in range(_time_signature.numerator):
                beat = QTargetBeat(
                    beatspan=beatspan,
                    offset_in_ms=current_offset_in_ms,
                    search_tree=search_tree,
                    tempo=tempo,
                )
                beats.append(beat)
                current_offset_in_ms += beatspan_duration_in_ms
        self._beats = tuple(beats)
        self._offset_in_ms = offset_in_ms
        self._search_tree = search_tree
        self._tempo = tempo
        self._time_signature = _time_signature
        self._use_full_measure = use_full_measure

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(offset_in_ms={self.offset_in_ms!r},"
        string += f" search_tree={self.search_tree!r}, tempo={self.tempo!r},"
        string += f" use_full_measure={self.use_full_measure!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def beats(self) -> tuple:
        """
        Gets the tuple of ``QTargetBeats`` contained by the ``QTargetMeasure``.

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> for q_target_beat in q_target_measure.beats:
            ...     q_target_beat.offset_in_ms
            ...
            Offset((1000, 1))
            Offset((2000, 1))
            Offset((3000, 1))
            Offset((4000, 1))

        """
        return self._beats

    @property
    def duration_in_ms(self) -> abjad.Duration:
        """
        Gets duration in milliseconds of the ``QTargetMeasure``:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> q_target_measure.duration_in_ms
            Duration(4000, 1)

        """
        return self.tempo.duration_to_milliseconds(self.time_signature.duration)

    @property
    def offset_in_ms(self) -> abjad.Offset:
        """
        Gets offset in milliseconds of the ``QTargetMeasure``:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> q_target_measure.offset_in_ms
            Offset((1000, 1))

        """
        return self._offset_in_ms

    @property
    def search_tree(self) -> _searchtrees.SearchTree:
        """
        Gets the search tree of ``QTargetMeasure``:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> q_target_measure.search_tree
            UnweightedSearchTree(definition={2: None})

        """
        return self._search_tree

    @property
    def tempo(self) -> abjad.MetronomeMark:
        r"""
        Gets the tempo of ``QTargetMeasure``:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> abjad.lilypond(q_target_measure.tempo)
            '\\tempo 4=60'

        """
        return self._tempo

    @property
    def time_signature(self) -> abjad.TimeSignature:
        """
        Gets the time signature of the ``QTargetMeasure``:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> q_target_measure.time_signature
            TimeSignature(pair=(4, 4), hide=False, partial=None)

        """
        return self._time_signature

    @property
    def use_full_measure(self) -> bool:
        """
        Is true when ``QTargetMeasure`` uses full measures:

        ..  container:: example

            >>> search_tree = nauert.UnweightedSearchTree({2: None})
            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
            >>> time_signature = abjad.TimeSignature((4, 4))

            >>> q_target_measure = nauert.QTargetMeasure(
            ...     offset_in_ms=abjad.Offset(1000),
            ...     search_tree=search_tree,
            ...     tempo=tempo,
            ...     time_signature=time_signature,
            ... )

            >>> q_target_measure.use_full_measure
            False

        """
        return self._use_full_measure



import abjad
import pytest

import nauert


def test_MeasurewiseQSchemaItem___new___01():
    item = nauert.MeasurewiseQSchemaItem()
    assert item.beatspan is None
    assert item.search_tree is None
    assert item.tempo is None
    assert item.time_signature is None


def test_MeasurewiseQSchemaItem___new___02():
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 60)
    item = nauert.MeasurewiseQSchemaItem(tempo=metronome_mark)
    assert item.beatspan is None
    assert item.search_tree is None
    assert item.tempo == metronome_mark
    assert item.time_signature is None


def test_MeasurewiseQSchemaItem___new___03():
    time_signature = abjad.TimeSignature((6, 8))
    item = nauert.MeasurewiseQSchemaItem(time_signature=time_signature)
    assert item.beatspan == abjad.Duration(1, 8)
    assert item.search_tree is None
    assert item.tempo is None
    assert item.time_signature == time_signature


def test_MeasurewiseQSchemaItem___new___04():
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 57)
    time_signature = abjad.TimeSignature((6, 8))
    item = nauert.MeasurewiseQSchemaItem(
        tempo=metronome_mark, time_signature=time_signature
    )
    assert item.beatspan == abjad.Duration(1, 8)
    assert item.search_tree is None
    assert item.tempo == metronome_mark
    assert item.time_signature == time_signature


def test_MeasurewiseQSchemaItem___new___05():
    tempo = abjad.MetronomeMark(textual_indication="lento")
    with pytest.raises(AssertionError):
        nauert.MeasurewiseQSchemaItem(tempo=tempo)



import abjad

import nauert


class Job(nauert.QuantizationJob):

    def __init__(self, number):
        self.number = number

    def __call__(self):
        self.result = [
            x for x in abjad.math.yield_all_compositions_of_integer(self.number)
        ]

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.number)


def test_SerialJobHandler___call___01():
    jobs = [Job(x) for x in range(1, 11)]
    job_handler = nauert.SerialJobHandler()
    job_handler(jobs)



import abc
import copy

import abjad

from . import qgrid as _qgrid


class SearchTree(abc.ABC):
    """
    Abstract search tree.

    ``SearchTrees`` encapsulate strategies for generating collections of
    ``QGrids``, given a set of ``QEventProxy`` instances as input.

    They allow composers to define the degree and quality of nested rhythmic
    subdivisions in the quantization output.  That is to say, they allow
    composers to specify what sorts of tuplets and ratios of pulses may be
    contained within other tuplets, to arbitrary levels of nesting.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_definition",)

    ### INITIALIZER ###

    def __init__(self, definition: dict | None = None):
        if definition is None:
            definition = self.default_definition
        else:
            assert self._is_valid_definition(definition)
        self._definition = definition

    ### SPECIAL METHODS ###

    def __call__(self, q_grid: _qgrid.QGrid) -> list[_qgrid.QGrid]:
        """
        Calls search tree.
        """
        assert isinstance(q_grid, _qgrid.QGrid)
        new_q_grids = []
        commands = self._generate_all_subdivision_commands(q_grid)
        for command in commands:
            new_q_grid = copy.deepcopy(q_grid)
            q_events = new_q_grid.subdivide_leaves(command)
            new_q_grid.fit_q_events(q_events)
            new_q_grids.append(new_q_grid)
        return new_q_grids

    def __eq__(self, argument) -> bool:
        """
        Is true when `argument` is a search tree with definition equal to that of
        this search tree. Otherwise false.
        """
        if type(self) is type(argument):
            if self.definition == argument.definition:
                return True
        return False

    def __hash__(self) -> int:
        """
        Hashes search tree.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super(SearchTree, self).__hash__()

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return f"{type(self).__name__}(definition={self.definition!r})"

    ### PRIVATE METHODS ###

    def _find_divisible_leaf_indices_and_subdivisions(
        self, q_grid: _qgrid.QGrid
    ) -> tuple[list[int], list[tuple[tuple[int, ...], ...]]]:
        # TODO: This should actually check for all QEvents which fall
        # within the leaf's duration,
        # including QEvents attached to the next leaf
        # It may be prudent to actually store QEvents in two lists:
        # before_offset and after_offset
        indices, subdivisions = [], []
        leaves = list(q_grid.leaves)
        i = 0
        for leaf_one, leaf_two in abjad.sequence.nwise(leaves):
            if leaf_one.is_divisible:
                succeeding_proxies = leaf_one.succeeding_q_event_proxies
                preceding_proxies = leaf_two.preceding_q_event_proxies
                if not preceding_proxies and all(
                    proxy.offset == leaf_one.start_offset
                    for proxy in succeeding_proxies
                ):
                    # proxies align perfectly with this leaf
                    pass
                elif preceding_proxies or succeeding_proxies:
                    parentage_ratios = leaf_one._get_parentage_ratios()
                    leaf_subdivisions = self._find_leaf_subdivisions(parentage_ratios)
                    if leaf_subdivisions:
                        indices.append(i)
                        subdivisions.append(tuple(leaf_subdivisions))
            i += 1
        return indices, subdivisions

    @abc.abstractmethod
    def _find_leaf_subdivisions(
        self, parentage_ratios: tuple
    ) -> tuple[tuple[int, ...], ...]:
        raise NotImplementedError

    def _generate_all_subdivision_commands(
        self, q_grid: _qgrid.QGrid
    ) -> tuple[tuple[tuple[int, tuple[int, int]], ...], ...]:
        indices, subdivisions = self._find_divisible_leaf_indices_and_subdivisions(
            q_grid
        )
        if not indices:
            return ()
        combinations = abjad.enumerate.outer_product(subdivisions)
        combinations = [tuple(_) for _ in combinations]
        return tuple(tuple(zip(indices, combo)) for combo in combinations)

    @abc.abstractmethod
    def _is_valid_definition(self, definition: dict) -> bool:
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def default_definition(self) -> dict:
        """
        Gets default search tree definition.
        """
        raise NotImplementedError

    @property
    def definition(self) -> dict:
        """
        Gets search tree definition.
        """
        return self._definition


class UnweightedSearchTree(SearchTree):
    r"""
    Unweighted search tree based on Paul Nauert's model.

    ..  container:: example

        >>> import pprint
        >>> search_tree = nauert.UnweightedSearchTree()
        >>> pprint.pprint(search_tree.definition)
        {2: {2: {2: {2: None}, 3: None}, 3: None, 5: None, 7: None},
            3: {2: {2: None}, 3: None, 5: None},
            5: {2: None, 3: None},
            7: {2: None},
            11: None,
            13: None}

    ..  container:: example

        The search tree defines how nodes in a ``QGrid`` may be subdivided, if
        they happen to contain ``QEvents`` (or, in actuality, ``QEventProxy``
        instances which reference ``QEvents``, but rescale their offsets
        between ``0`` and ``1``).

        In the default definition, the root node of the ``QGrid`` may be
        subdivided into ``2``, ``3``, ``5``, ``7``, ``11`` or ``13`` equal
        parts. If divided into ``2`` parts, the divisions of the root node may
        be divided again into ``2``, ``3``, ``5`` or ``7``, and so forth.

        This definition is structured as a collection of nested dictionaries,
        whose keys are integers, and whose values are either the sentinel
        ``None`` indicating no further permissable divisions, or dictionaries
        obeying these same rules, which then indicate the possibilities for
        further division.

        Calling a ``UnweightedSearchTree`` with a ``QGrid`` instance will
        generate all permissable subdivided ``QGrids``, according to the
        definition of the called search tree:

        >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(130), [0, 1, 4])
        >>> q_event_b = nauert.PitchedQEvent(abjad.Offset(150), [2, 3, 5])
        >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.5))
        >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.667))
        >>> q_grid = nauert.QGrid()
        >>> q_grid.fit_q_events([proxy_a, proxy_b])
        >>> q_grids = search_tree(q_grid)
        >>> for grid in q_grids:
        ...     print(grid.rtm_format)
        (1 (1 1))
        (1 (1 1 1))
        (1 (1 1 1 1 1))
        (1 (1 1 1 1 1 1 1))
        (1 (1 1 1 1 1 1 1 1 1 1 1))
        (1 (1 1 1 1 1 1 1 1 1 1 1 1 1))

    ..  container:: example

        A custom ``UnweightedSearchTree`` may be defined by passing in a
        dictionary, as described above. The following search tree only permits
        divisions of the root node into ``2s`` and ``3s``, and if divided into
        ``2``, a node may be divided once more into ``2`` parts:

        >>> definition = {2: {2: None}, 3: None}
        >>> search_tree = nauert.UnweightedSearchTree(definition)

        >>> q_grids = search_tree(q_grid)
        >>> for grid in q_grids:
        ...     print(grid.rtm_format)
        (1 (1 1))
        (1 (1 1 1))

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### PRIVATE METHODS ###

    def _find_leaf_subdivisions(
        self, parentage_ratios: tuple
    ) -> tuple[tuple[int, ...], ...]:
        parentage = [x[1] for x in parentage_ratios[1:]]
        if not parentage:
            return tuple((1,) * x for x in sorted(self._definition.keys()))
        node = self._definition[parentage[0]]
        for item in parentage[1:]:
            node = node[item]
            if node is None:
                return ()
        if node is None:
            return ()
        return tuple((1,) * x for x in sorted(node.keys()))

    def _is_valid_definition(self, definition: dict) -> bool:
        def recurse(n):
            results = []
            for key in n:
                if (
                    not isinstance(key, int)
                    or not 0 < key
                    or not abjad.math.divisors(key) == [1, key]
                ):
                    results.append(False)
                elif not isinstance(n[key], (dict, type(None))):
                    results.append(False)
                elif isinstance(n[key], dict) and not recurse(n[key]):
                    results.append(False)
                else:
                    results.append(True)
            return results

        if not isinstance(definition, dict) or not len(definition):
            return False
        return all(recurse(definition))

    ### PUBLIC PROPERTIES ###

    @property
    def default_definition(self) -> dict:
        """
        The default search tree definition, based on the search tree given by
        Paul Nauert:

        >>> import pprint
        >>> search_tree = nauert.UnweightedSearchTree()
        >>> pprint.pprint(search_tree.default_definition)
        {2: {2: {2: {2: None}, 3: None}, 3: None, 5: None, 7: None},
            3: {2: {2: None}, 3: None, 5: None},
            5: {2: None, 3: None},
            7: {2: None},
            11: None,
            13: None}

        """
        return {
            2: {  # 1/2
                2: {2: {2: None}, 3: None},  # 1/4  # 1/8  # 1/16  # 1/12
                3: None,  # 1/6
                5: None,  # 1/10
                7: None,  # 1/14
            },
            3: {2: {2: None}, 3: None, 5: None},  # 1/3  # 1/6  # 1/12  # 1/9  # 1/15
            5: {2: None, 3: None},  # 1/5  # 1/10  # 1/15
            7: {2: None},  # 1/7  # 1/14
            11: None,  # 1/11
            13: None,  # 1/13
        }


class WeightedSearchTree(SearchTree):
    r"""
    Weighted search tree.

    ..  container:: example

        Allows for dividing nodes in a q-grid into parts with unequal weights.

        >>> search_tree = nauert.WeightedSearchTree()
        >>> search_tree.definition
        {'divisors': (2, 3, 5, 7), 'max_depth': 3, 'max_divisions': 2}

    ..  container:: example

        In ``WeightedSearchTree``'s definition:

            * ``divisors`` controls the sum of the parts of the ratio a node
                may be divided into,
            * ``max_depth`` controls how many levels of tuplet nesting
                are permitted, and
            * ``max_divisions`` controls the maximum permitted length of the
                weights in the ratio.

        Thus, the default ``WeightedSearchTree`` permits the following ratios:

        >>> for composition in search_tree.all_compositions:
        ...     composition
        ...
        (1, 1)
        (2, 1)
        (1, 2)
        (4, 1)
        (3, 2)
        (2, 3)
        (1, 4)
        (6, 1)
        (5, 2)
        (4, 3)
        (3, 4)
        (2, 5)
        (1, 6)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_all_compositions", "_compositions", "_definition")

    ### INITIALIZER ###

    def __init__(self, definition: dict | None = None):
        SearchTree.__init__(self, definition)
        self._compositions = self._precompute_compositions()
        all_compositions = []
        for value in list(self._compositions.values()):
            all_compositions.extend(value)
        self._all_compositions = tuple(all_compositions)

    ### PRIVATE METHODS ###

    def _find_leaf_subdivisions(
        self, parentage_ratios: tuple
    ) -> tuple[tuple[int, ...], ...]:
        if len(parentage_ratios[1:]) < self._definition["max_depth"]:
            return self._all_compositions
        return ()

    def _is_valid_definition(self, definition: dict) -> bool:
        if not isinstance(definition, dict):
            return False
        elif "divisors" not in definition:
            return False
        elif not len(definition["divisors"]):
            return False
        elif not all(isinstance(x, int) and 1 < x for x in definition["divisors"]):
            return False
        elif not all(abjad.math.divisors(x) == [1, x] for x in definition["divisors"]):
            return False
        elif "max_depth" not in definition:
            return False
        elif not isinstance(definition["max_depth"], int):
            return False
        elif not 0 < definition["max_depth"]:
            return False
        elif "max_divisions" not in definition:
            return False
        elif not isinstance(definition["max_divisions"], int):
            return False
        elif not 1 < definition["max_divisions"]:
            return False
        return True

    def _precompute_compositions(
        self,
    ) -> dict[int, list[tuple[int, ...]]]:
        compositions = {}
        max_divisions = self._definition["max_divisions"]
        for divisor in self._definition["divisors"]:
            compositions[divisor] = [
                tuple(x)
                for x in abjad.math.yield_all_compositions_of_integer(divisor)
                if 1 < len(x) <= max_divisions
            ]
        return compositions

    ### PUBLIC PROPERTIES ###

    @property
    def all_compositions(self) -> tuple[tuple[int, ...], ...]:
        """
        Gets all compositions of weighted search tree.
        """
        return self._all_compositions

    @property
    def default_definition(self) -> dict:
        """
        Gets default definition of weighted search tree.
        """
        return {"divisors": (2, 3, 5, 7), "max_depth": 3, "max_divisions": 2}



import copy

import abjad

import nauert


def test_QGridLeaf___copy___01():
    leaf = nauert.QGridLeaf(abjad.Duration(1))
    copied = copy.deepcopy(leaf)
    assert format(leaf) == format(copied)
    assert leaf != copied
    assert leaf is not copied


def test_QGridLeaf___copy___02():
    sqe = nauert.SilentQEvent(abjad.Offset(1000))
    leaf = nauert.QGridLeaf(
        abjad.Duration(2),
        [nauert.QEventProxy(sqe, abjad.Offset(0.5))],
    )
    copied = copy.deepcopy(leaf)
    assert format(leaf) == format(copied)
    assert leaf != copied
    assert leaf is not copied



import collections
import copy
import itertools
import numbers
import typing

import abjad

from . import qevents as _qevents


class QEventSequence:
    r"""
    Q-event sequence.

    Contains only pitched q-events and silent q-events, and terminates with a
    single terminal q-event.

    A q-event sequence is the primary input to the quantizer.

    ..  container:: example

        A q-event sequence provides a number of convenience functions to assist
        with instantiating new sequences:

        >>> durations = (1000, -500, 1250, -500, 750)
        >>> sequence = nauert.QEventSequence.from_millisecond_durations(durations)
        >>> for q_event in sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_sequence",)

    ### INITIALIZER ###

    def __init__(self, sequence):
        q_event_classes = (_qevents.PitchedQEvent, _qevents.SilentQEvent)
        if sequence is None:
            self._sequence = ()
            return
        else:
            assert 1 < len(sequence)
            assert all(
                isinstance(q_event, q_event_classes) for q_event in sequence[:-1]
            )
            assert isinstance(sequence[-1], _qevents.TerminalQEvent)
            offsets = [x.offset for x in sequence]
            offset_sequence = list(offsets)
            assert abjad.sequence.is_increasing(offset_sequence, strict=False)
            assert 0 <= sequence[0].offset
            self._sequence = tuple(sequence)

    ### SPECIAL METHODS ###

    def __contains__(self, argument) -> bool:
        """
        Is true when q-event sequence contains ``argument``. Otherwise false.
        """
        return argument in self._sequence

    def __eq__(self, argument) -> bool:
        """
        Is true when q-event sequence equals ``argument``. Otherwise false.
        """
        if type(self) is type(argument):
            if self.sequence == argument.sequence:
                return True
        return False

    @typing.overload
    def __getitem__(self, argument: int) -> _qevents.QEvent:
        pass

    @typing.overload
    def __getitem__(self, argument: slice) -> tuple[_qevents.QEvent, ...]:
        pass

    def __getitem__(
        self, argument: int | slice
    ) -> _qevents.QEvent | tuple[_qevents.QEvent, ...]:
        """
        Gets item or slice identified by `argument`.

        Returns item or slice.
        """
        return self._sequence.__getitem__(argument)

    def __hash__(self) -> int:
        """
        Hashes q-event sequence.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super(QEventSequence, self).__hash__()

    def __iter__(self) -> typing.Iterator[_qevents.QEvent]:
        """
        Iterates q-event sequence.
        """
        for x in self._sequence:
            yield x

    def __len__(self) -> int:
        """
        Gets length of q-event sequence.
        """
        return len(self._sequence)

    ### PUBLIC PROPERTIES ###

    @property
    def duration_in_ms(self) -> abjad.Duration:
        r"""
        Get duration ``QEventSequence`` in milliseconds:

        >>> durations = (1000, -500, 1250, -500, 750)
        >>> sequence = nauert.QEventSequence.from_millisecond_durations(durations)
        >>> sequence.duration_in_ms
        Duration(4000, 1)

        """
        return abjad.Duration(self[-1].offset)

    @property
    def sequence(self) -> tuple:
        r"""
        Gets sequence of q-events.

        >>> durations = (1000, -500, 1250, -500, 750)
        >>> sequence = nauert.QEventSequence.from_millisecond_durations(durations)
        >>> for q_event in sequence.sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

        """
        return self._sequence

    ### PUBLIC METHODS ###

    @classmethod
    def from_millisecond_durations(
        class_,
        milliseconds: typing.Sequence[int | float],
        fuse_silences: bool = False,
    ) -> "QEventSequence":
        r"""
        Changes sequence of millisecond ``durations`` to ``QEventSequence``:

        >>> durations = [-250, 500, -1000, 1250, -1000]
        >>> sequence = nauert.QEventSequence.from_millisecond_durations(durations)
        >>> for q_event in sequence:
        ...     q_event
        ...
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        TerminalQEvent(...)

        """
        durations: typing.Sequence[int | float]
        if fuse_silences:
            durations = [
                _ for _ in abjad.sequence.sum_by_sign(milliseconds, sign=[-1]) if _
            ]
        else:
            durations = milliseconds
        sums = abjad.math.cumulative_sums([abs(_) for _ in durations])
        offsets = [abjad.Offset(_) for _ in sums]
        q_events: list[_qevents.QEvent] = []
        for offset, duration in zip(offsets, durations):
            q_event: _qevents.QEvent
            if duration < 0:
                q_event = _qevents.SilentQEvent(offset)
            else:
                q_event = _qevents.PitchedQEvent(offset, [0])
            q_events.append(q_event)
        q_event = _qevents.TerminalQEvent(offsets[-1])
        q_events.append(q_event)
        return class_(q_events)

    @classmethod
    def from_millisecond_offsets(
        class_, offsets: typing.Sequence[abjad.Offset]
    ) -> "QEventSequence":
        r"""
        Changes millisecond ``offsets`` to ``QEventSequence``:

        >>> numbers = [0, 250, 750, 1750, 3000, 4000]
        >>> offsets = [abjad.Offset(_) for _ in numbers]
        >>> sequence = nauert.QEventSequence.from_millisecond_offsets(offsets)
        >>> for q_event in sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        PitchedQEvent(...)
        PitchedQEvent(...)
        PitchedQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

        """
        assert all(isinstance(_, abjad.Offset) for _ in offsets), repr(offsets)
        q_events: list[_qevents.QEvent] = []
        q_events.extend([_qevents.PitchedQEvent(_, [0]) for _ in offsets[:-1]])
        q_events.append(_qevents.TerminalQEvent(offsets[-1]))
        return class_(q_events)

    @classmethod
    def from_millisecond_pitch_attachment_tuples(
        class_, tuples: typing.Iterable[tuple]
    ) -> "QEventSequence":
        r"""
        Changes (millisecond-duration, pitch, attachment) ``tuples`` into
        ``QEventSequence``:

        >>> durations = [250, 500, 1000, 1250, 1000]
        >>> pitches = [(0,), None, (2, 3), None, (1,)]
        >>> attachments = [("foo",), (), (), (), ("foobar", "foo")]
        >>> tuples = tuple(zip(durations, pitches, attachments))
        >>> method = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples
        >>> sequence = method(tuples)
        >>> for q_event in sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

        """
        assert isinstance(tuples, collections.abc.Iterable)
        assert all(isinstance(_, collections.abc.Iterable) for _ in tuples)
        assert all(len(_) == 3 for _ in tuples)
        assert all(0 < duration for duration, _, _ in tuples)
        pitch_prototype = (numbers.Number, type(None), collections.abc.Sequence)
        for _, pitches, attachments in tuples:
            if attachments is None:
                raise Exception(attachments)
            assert isinstance(pitches, pitch_prototype)
            if isinstance(pitches, collections.abc.Sequence):
                assert 0 < len(pitches)
                assert all(isinstance(_, numbers.Number) for _ in pitches)
            if pitches is None:
                assert not attachments, repr(attachments)
        # fuse silences
        g = itertools.groupby(tuples, lambda _: _[1] is not None)
        groups = []
        for value, group in g:
            if value:
                groups.extend(list(group))
            else:
                duration = sum(_[0] for _ in group)
                groups.append((duration, None, ()))
        # find offsets
        offsets = abjad.math.cumulative_sums([abs(_[0]) for _ in groups])
        offsets = [abjad.Offset(_) for _ in offsets]
        # build QEvents
        q_events = [
            _qevents.QEvent.from_offset_pitches_attachments(
                offset, pitches, attachments
            )
            for offset, (_, pitches, attachments) in zip(offsets, groups)
        ]
        q_events.append(_qevents.TerminalQEvent(offsets[-1]))
        return class_(q_events)

    @classmethod
    def from_millisecond_pitch_pairs(
        class_, pairs: typing.Iterable[tuple]
    ) -> "QEventSequence":
        r"""
        Changes (millisecond-duration, pitch) ``pairs`` into ``QEventSequence``:

        >>> durations = [250, 500, 1000, 1250, 1000]
        >>> pitches = [(0,), None, (2, 3), None, (1,)]
        >>> pairs = tuple(zip(durations, pitches))
        >>> sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> for q_event in sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

        """
        assert isinstance(pairs, collections.abc.Iterable)
        assert all(isinstance(_, collections.abc.Iterable) for _ in pairs)
        assert all(len(_) == 2 for _ in pairs)
        assert all(0 < duration for duration, _ in pairs)
        pitch_prototype = (numbers.Number, type(None), collections.abc.Sequence)
        for _, pitches in pairs:
            assert isinstance(pitches, pitch_prototype)
            if isinstance(pitches, collections.abc.Sequence):
                assert 0 < len(pitches)
                assert all(isinstance(_, numbers.Number) for _ in pitches)
        return class_.from_millisecond_pitch_attachment_tuples(
            [(duration, pitches, ()) for duration, pitches in pairs]
        )

    @classmethod
    def from_tempo_scaled_durations(
        class_,
        durations: typing.Sequence[abjad.Duration],
        tempo: abjad.MetronomeMark,
    ) -> "QEventSequence":
        r"""
        Changes ``durations``, scaled by ``tempo`` into a ``QEventSequence``:

        ..  container:: example

            >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 174)
            >>> pairs = [(1, 4), (-3, 16), (1, 16), (-1, 2)]
            >>> durations = [abjad.Duration(_) for _ in pairs]
            >>> method = nauert.QEventSequence.from_tempo_scaled_durations
            >>> sequence = method(durations, tempo=tempo)
            >>> for q_event in sequence:
            ...     q_event
            ...
            PitchedQEvent(...)
            SilentQEvent(...)
            PitchedQEvent(...)
            SilentQEvent(...)
            TerminalQEvent(...)

        """
        assert all(isinstance(_, abjad.Duration) for _ in durations), repr(durations)
        assert isinstance(tempo, abjad.MetronomeMark)
        durations = [x for x in abjad.sequence.sum_by_sign(durations, sign=[-1]) if x]
        durations = [tempo.duration_to_milliseconds(_) for _ in durations]
        offsets = abjad.math.cumulative_sums([abs(_) for _ in durations])
        offsets = [abjad.Offset(_) for _ in offsets]
        q_events = []
        for offset, duration in zip(offsets, durations):
            offset = abjad.Offset(offset)
            assert isinstance(duration, abjad.Duration)
            q_event: _qevents.QEvent
            # negative duration indicates silence
            if duration < 0:
                q_event = _qevents.SilentQEvent(offset)
            # otherwise use middle C
            else:
                q_event = _qevents.PitchedQEvent(offset, [0])
            q_events.append(q_event)
        # insert terminating silence _qevents.QEvent
        q_events.append(_qevents.TerminalQEvent(offsets[-1]))
        return class_(q_events)

    @classmethod
    def from_tempo_scaled_leaves(class_, leaves, tempo=None) -> "QEventSequence":
        r"""
        Changes ``leaves``, optionally with ``tempo`` into a ``QEventSequence``:

        >>> staff = abjad.Staff("c'4 <d' fs'>8. r16 gqs'2")
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 72)
        >>> sequence = nauert.QEventSequence.from_tempo_scaled_leaves(staff[:], tempo)
        >>> for q_event in sequence:
        ...     q_event
        ...
        PitchedQEvent(...)
        PitchedQEvent(...)
        SilentQEvent(...)
        PitchedQEvent(...)
        TerminalQEvent(...)

        If ``tempo`` is ``None``, all leaves in ``leaves`` must have an
        effective, non-imprecise tempo. The millisecond-duration of each leaf
        will be determined by its effective tempo.
        """
        assert len(leaves)
        match tempo:
            case None:
                prototype = abjad.MetronomeMark
                assert abjad.get.effective(leaves[0], prototype) is not None
            case abjad.MetronomeMark():
                tempo = copy.deepcopy(tempo)
            case tuple():
                tempo = abjad.MetronomeMark(*tempo)
            case _:
                raise TypeError(tempo)
        # sort by silence and tied leaves
        groups = []
        for rvalue, rgroup in itertools.groupby(
            leaves, lambda x: isinstance(x, (abjad.Rest, abjad.Skip))
        ):
            if rvalue:
                groups.append(list(rgroup))
            else:
                for tvalue, tgroup in itertools.groupby(
                    rgroup, lambda x: abjad._iterlib._get_logical_tie_leaves(x)
                ):
                    groups.append(list(tgroup))
        # calculate lists of pitches and durations
        durations = []
        pitches = []
        for group in groups:
            # get millisecond cumulative duration
            if tempo is not None:
                duration = sum(
                    tempo.duration_to_milliseconds(x._get_duration()) for x in group
                )
            else:
                duration = sum(
                    abjad.get.effective(
                        x, abjad.MetronomeMark
                    ).duration_to_milliseconds(x._get_duration())
                    for x in group
                )
            durations.append(duration)
            pitch: typing.Any
            # get pitch of first leaf in group
            if isinstance(group[0], (abjad.Rest, abjad.Skip)):
                pitch = None
            elif isinstance(group[0], abjad.Note):
                assert group[0].written_pitch is not None
                pitch = group[0].written_pitch.number
            # chord
            else:
                pitch = [x.written_pitch.number for x in group[0].note_heads]
            pitches.append(pitch)
        # convert durations and pitches to QEvents and return
        return class_.from_millisecond_pitch_pairs(tuple(zip(durations, pitches)))



import abjad

import nauert


def test_MeasurewiseAttackPointOptimizer___call___01():
    string = r"""
        \tuplet 11/8
        {
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            ~
            a''8
            \grace {
                e''16
                e''16
            }
            d''8
            ~
            d''8
            r8
            r8
        }
        """
    string = abjad.string.normalize(string)
    container = abjad.Container(string)
    time_signature = abjad.TimeSignature((4, 4))
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    attack_point_optimizer(container, time_signature)
    string = abjad.lilypond(container)
    assert string == abjad.string.normalize(
        r"""
        {
            \tuplet 11/8
            {
                a''2..
                \grace {
                    e''16
                    e''16
                }
                d''4
                r4
            }
        }
        """
    ), print(string)



import abjad

import nauert


def test_QEventSequence_from_tempo_scaled_durations_01():
    """
    Test basic functionality.
    """
    durations = [abjad.Duration(x) for x in [(1, 4), (1, 3), (1, 7), (2, 5), (3, 4)]]
    tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 55)
    q_events = nauert.QEventSequence.from_tempo_scaled_durations(durations, tempo)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(12000, 11), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(28000, 11), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(244000, 77), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(34400, 7), (abjad.NamedPitch("c'"),)),
            nauert.TerminalQEvent(abjad.Offset(630400, 77)),
        )
    )


def test_QEventSequence_from_tempo_scaled_durations_02():
    """
    Silences are fused.
    """
    durations = [
        abjad.Duration(x)
        for x in [(1, 4), (-1, 4), (1, 4), (1, 4), (-1, 4), (-1, 4), (1, 4)]
    ]
    tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 77)
    q_events = nauert.QEventSequence.from_tempo_scaled_durations(durations, tempo)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0, 1), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(60000, 77)),
            nauert.PitchedQEvent(abjad.Offset(120000, 77), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(180000, 77), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(240000, 77)),
            nauert.PitchedQEvent(abjad.Offset(360000, 77), (abjad.NamedPitch("c'"),)),
            nauert.TerminalQEvent(abjad.Offset(60000, 11)),
        )
    )



import abjad

import nauert


def test_MeasurewiseQSchema___getitem___01():
    schema = nauert.MeasurewiseQSchema()
    assert (
        schema[0]
        == schema[1]
        == schema[2]
        == {
            "search_tree": nauert.UnweightedSearchTree(),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 60),
            "time_signature": abjad.TimeSignature((4, 4)),
            "use_full_measure": False,
        }
    )


def test_MeasurewiseQSchema___getitem___02():
    item_a = nauert.MeasurewiseQSchemaItem(
        search_tree=nauert.UnweightedSearchTree({2: None})
    )
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 76)
    item_b = nauert.MeasurewiseQSchemaItem(tempo=metronome_mark)
    time_signature = abjad.TimeSignature((3, 4))
    item_c = nauert.MeasurewiseQSchemaItem(time_signature=time_signature)
    item_d = nauert.MeasurewiseQSchemaItem(
        search_tree=nauert.UnweightedSearchTree({5: None}),
        use_full_measure=True,
    )
    schema = nauert.MeasurewiseQSchema(
        {2: item_a, 4: item_b, 7: item_c, 8: item_d},
        search_tree=nauert.UnweightedSearchTree({3: None}),
        tempo=abjad.MetronomeMark(abjad.Duration(1, 8), 58),
        time_signature=(5, 8),
        use_full_measure=False,
    )
    assert (
        schema[0]
        == schema[1]
        == {
            "search_tree": nauert.UnweightedSearchTree({3: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 8), 58),
            "time_signature": abjad.TimeSignature((5, 8)),
            "use_full_measure": False,
        }
    )
    assert (
        schema[2]
        == schema[3]
        == {
            "search_tree": nauert.UnweightedSearchTree({2: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 8), 58),
            "time_signature": abjad.TimeSignature((5, 8)),
            "use_full_measure": False,
        }
    )
    assert (
        schema[4]
        == schema[5]
        == schema[6]
        == {
            "search_tree": nauert.UnweightedSearchTree({2: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 76),
            "time_signature": abjad.TimeSignature((5, 8)),
            "use_full_measure": False,
        }
    )
    assert schema[7] == {
        "search_tree": nauert.UnweightedSearchTree({2: None}),
        "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 76),
        "time_signature": abjad.TimeSignature((3, 4)),
        "use_full_measure": False,
    }
    assert (
        schema[8]
        == schema[9]
        == schema[1000]
        == {
            "search_tree": nauert.UnweightedSearchTree({5: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 76),
            "time_signature": abjad.TimeSignature((3, 4)),
            "use_full_measure": True,
        }
    )



import nauert


def test_UnweightedSearchTree__find_leaf_subdivisions_01():
    definition = {2: {2: {2: None}, 3: None}, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    assert search_tree._find_leaf_subdivisions((1, (1, 2))) == (
        (1, 1),
        (1, 1, 1),
    )
    assert search_tree._find_leaf_subdivisions((1, (1, 2), (1, 2))) == ((1, 1),)
    assert search_tree._find_leaf_subdivisions((1, (1, 2), (1, 2), (1, 2))) == ()
    assert search_tree._find_leaf_subdivisions((1, (1, 2), (1, 3))) == ()
    assert search_tree._find_leaf_subdivisions((1, (1, 5))) == ()



import abc

from . import qgrid as _qgrid
from . import qtargetitems as _qtargetitems


class Heuristic(abc.ABC):
    """
    Abstract heuristic.

    Heuristics rank Q-grids according to the criteria they encapsulate.

    Heuristics provide the means by which the quantizer selects a single
    ``QGrid`` from all computed ``QGrids`` for any given ``QTargetBeat`` to
    represent that beat.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        pass

    ### SPECIAL METHODS ###

    def __call__(
        self, q_target_beats: tuple[_qtargetitems.QTargetBeat, ...]
    ) -> tuple[_qtargetitems.QTargetBeat, ...]:
        """
        Calls heuristic.
        """
        assert len(q_target_beats)
        assert all(isinstance(x, _qtargetitems.QTargetBeat) for x in q_target_beats)
        return self._process(q_target_beats)

    ### PRIVATE METHODS ###

    @abc.abstractmethod
    def _process(
        self, q_target_beats: tuple[_qtargetitems.QTargetBeat, ...]
    ) -> tuple[_qtargetitems.QTargetBeat, ...]:
        raise NotImplementedError


class DistanceHeuristic(Heuristic):
    r"""
    Distance heuristic.

    Considers only the computed distance of each ``QGrid`` and the number of
    leaves of that ``QGrid`` when choosing the optimal ``QGrid`` for a given
    ``QTargetBeat``.

    The ``QGrid`` with the smallest distance and fewest number of leaves will
    be selected.

    ..  container:: example

        >>> durations = [1000] * 8
        >>> pitches = range(8)
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> heuristic = nauert.DistanceHeuristic()
        >>> voice = nauert.quantize(q_event_sequence, heuristic=heuristic)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    cs'4
                    d'4
                    ef'4
                }
                {
                    e'4
                    f'4
                    fs'4
                    g'4
                }
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### PRIVATE METHODS ###

    def _process(
        self, q_target_beats: tuple[_qtargetitems.QTargetBeat, ...]
    ) -> tuple[_qtargetitems.QTargetBeat, ...]:
        for q_target_beat in q_target_beats:
            q_grids = q_target_beat.q_grids
            if q_grids:
                sorted_q_grids = sorted(
                    q_grids, key=lambda x: (x.distance, len(x.leaves))
                )
                q_target_beat._q_grid = sorted_q_grids[0]
            else:
                q_target_beat._q_grid = _qgrid.QGrid()
        return q_target_beats



import abjad

import nauert


def test_QEventSequence_from_tempo_scaled_leaves_01():
    staff = abjad.Staff([])
    staff.append(abjad.Note(0, (1, 4)))
    staff.append(abjad.Rest((1, 4)))
    staff.append(abjad.Rest((1, 8)))
    staff.append(abjad.Note(1, (1, 8)))
    staff.append(abjad.Note(1, (1, 8)))
    staff.append(abjad.Note(2, (1, 8)))
    staff.append(abjad.Note(2, (1, 8)))
    staff.append(abjad.Note(3, (1, 8)))
    staff.append(abjad.Skip((1, 4)))
    staff.append(abjad.Rest((1, 4)))
    staff.append(abjad.Note(3, (1, 8)))
    staff.append(abjad.Chord([0, 1, 4], (1, 4)))
    abjad.tie(staff[3:5])
    abjad.tie(staff[5:7])
    tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 55)
    leaves = abjad.select.leaves(staff)
    q_events = nauert.QEventSequence.from_tempo_scaled_leaves(leaves, tempo)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0, 1), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(12000, 11)),
            nauert.PitchedQEvent(abjad.Offset(30000, 11), (abjad.NamedPitch("cs'"),)),
            nauert.PitchedQEvent(abjad.Offset(42000, 11), (abjad.NamedPitch("d'"),)),
            nauert.PitchedQEvent(abjad.Offset(54000, 11), (abjad.NamedPitch("ef'"),)),
            nauert.SilentQEvent(abjad.Offset(60000, 11)),
            nauert.PitchedQEvent(abjad.Offset(84000, 11), (abjad.NamedPitch("ef'"),)),
            nauert.PitchedQEvent(
                abjad.Offset(90000, 11),
                (
                    abjad.NamedPitch("c'"),
                    abjad.NamedPitch("cs'"),
                    abjad.NamedPitch("e'"),
                ),
            ),
            nauert.TerminalQEvent(abjad.Offset(102000, 11)),
        )
    )


def test_QEventSequence_from_tempo_scaled_leaves_02():
    staff = abjad.Staff([])
    staff.append(abjad.Note(0, (1, 4)))
    staff.append(abjad.Rest((1, 4)))
    staff.append(abjad.Rest((1, 8)))
    staff.append(abjad.Note(1, (1, 8)))
    staff.append(abjad.Note(1, (1, 8)))
    staff.append(abjad.Note(2, (1, 8)))
    staff.append(abjad.Note(2, (1, 8)))
    staff.append(abjad.Note(3, (1, 8)))
    staff.append(abjad.Skip((1, 4)))
    staff.append(abjad.Rest((1, 4)))
    staff.append(abjad.Note(3, (1, 8)))
    staff.append(abjad.Chord([0, 1, 4], (1, 4)))
    abjad.tie(staff[3:5])
    abjad.tie(staff[5:7])
    tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 58)
    abjad.attach(tempo, staff[0], context="Staff")
    tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 77)
    abjad.attach(tempo, staff[9], context="Staff")
    leaves = abjad.select.leaves(staff)
    q_events = nauert.QEventSequence.from_tempo_scaled_leaves(leaves)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0, 1), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(30000, 29)),
            nauert.PitchedQEvent(abjad.Offset(75000, 29), (abjad.NamedPitch("cs'"),)),
            nauert.PitchedQEvent(abjad.Offset(105000, 29), (abjad.NamedPitch("d'"),)),
            nauert.PitchedQEvent(abjad.Offset(135000, 29), (abjad.NamedPitch("ef'"),)),
            nauert.SilentQEvent(abjad.Offset(150000, 29)),
            nauert.PitchedQEvent(
                abjad.Offset(15600000, 2233), (abjad.NamedPitch("ef'"),)
            ),
            nauert.PitchedQEvent(
                abjad.Offset(16470000, 2233),
                (
                    abjad.NamedPitch("c'"),
                    abjad.NamedPitch("cs'"),
                    abjad.NamedPitch("e'"),
                ),
            ),
            nauert.TerminalQEvent(abjad.Offset(18210000, 2233)),
        )
    )



import copy

import abjad

import nauert


def test_QGridContainer___copy___01():
    tree = nauert.QGridContainer(
        (1, 1),
        children=[
            nauert.QGridLeaf(abjad.Duration(1, 1)),
            nauert.QGridContainer(
                (2, 1),
                children=[
                    nauert.QGridLeaf(abjad.Duration(3, 1)),
                    nauert.QGridContainer(
                        (4, 1),
                        children=[
                            nauert.QGridLeaf(abjad.Duration(1, 1)),
                            nauert.QGridLeaf(abjad.Duration(1, 1)),
                            nauert.QGridLeaf(abjad.Duration(1, 1)),
                        ],
                    ),
                ],
            ),
            nauert.QGridLeaf(abjad.Duration(2, 1)),
        ],
    )
    copied = copy.deepcopy(tree)
    assert format(tree) == format(copied)
    assert tree is not copied
    assert tree[0] is not copied[0]
    assert tree[1] is not copied[1]
    assert tree[2] is not copied[2]



import abjad

import nauert


def test_UnweightedSearchTree__find_divisible_leaf_indices_and_subdivisions_01():
    definition = {2: {2: {2: None}, 3: None}, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(0), ["A"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 5), ["B"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 4), ["C"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 3), ["D"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 5), ["E"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["F"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    g = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 5), ["G"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    h = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 3), ["H"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    i = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 4), ["I"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    j = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(4, 5), ["J"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    k = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1), ["K"]),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    q_grid.fit_q_events([a, b, c, d, e, f, g, h, i, j, k])
    indices, subdivisions = search_tree._find_divisible_leaf_indices_and_subdivisions(
        q_grid
    )
    assert indices == [0]
    assert subdivisions == [((1, 1), (1, 1, 1, 1, 1))]
